@@ -2,21 +2,20 @@
 Author       : zhangxianbing
 Date         : 2020-12-27 09:22:14
 LastEditors  : zhangxianbing
-LastEditTime : 2021-03-02 15:56:00
+LastEditTime : 2022-03-14 10:33:55
 Description  : JSONPath
 """
-__version__ = "1.0.4"
+__version__ = "1.0.6"
 __author__ = "zhangxianbing"
 
 import json
 import logging
 import os
 import re
+import sys
 from collections import defaultdict
 from typing import Union
 
-# pylint: disable=invalid-name,missing-function-docstring,missing-class-docstring,eval-used,logging-fstring-interpolation
-
 
 def create_logger(name: str = None, level: Union[int, str] = logging.INFO):
     """Get or create a logger used for local debug."""
@@ -57,6 +56,8 @@ class JSONPath:
     # save special patterns
     REP_GET_QUOTE = re.compile(r"['](.*?)[']")
     REP_PUT_QUOTE = re.compile(r"#Q(\d+)")
+    REP_GET_BACKQUOTE = re.compile(r"[`](.*?)[`]")
+    REP_PUT_BACKQUOTE = re.compile(r"#BQ(\d+)")
     REP_GET_BRACKET = re.compile(r"[\[](.*?)[\]]")
     REP_PUT_BRACKET = re.compile(r"#B(\d+)")
     REP_GET_PAREN = re.compile(r"[\(](.*?)[\)]")
@@ -70,11 +71,13 @@ class JSONPath:
     )
 
     # annotations
+    f: list
     segments: list
     lpath: int
     subx = defaultdict(list)
     result: list
     result_type: str
+    eval_func: callable
 
     def __init__(self, expr: str):
         expr = self._parse_expr(expr)
@@ -82,7 +85,9 @@ class JSONPath:
         self.lpath = len(self.segments)
         logger.debug(f"segments  : {self.segments}")
 
-    def parse(self, obj, result_type="VALUE"):
+        self.caller_globals = sys._getframe(1).f_globals
+
+    def parse(self, obj, result_type="VALUE", eval_func=eval):
         if not isinstance(obj, (list, dict)):
             raise TypeError("obj must be a list or a dict.")
 
@@ -91,22 +96,30 @@ class JSONPath:
                 f"result_type must be one of {tuple(JSONPath.RESULT_TYPE.keys())}"
             )
         self.result_type = result_type
+        self.eval_func = eval_func
 
         self.result = []
         self._trace(obj, 0, "$")
 
         return self.result
 
+    def search(self, obj, result_type="VALUE"):
+        return self.parse(obj, result_type)
+
     def _parse_expr(self, expr):
         logger.debug(f"before expr : {expr}")
-
+        # pick up special patterns
         expr = JSONPath.REP_GET_QUOTE.sub(self._get_quote, expr)
+        expr = JSONPath.REP_GET_BACKQUOTE.sub(self._get_backquote, expr)
         expr = JSONPath.REP_GET_BRACKET.sub(self._get_bracket, expr)
         expr = JSONPath.REP_GET_PAREN.sub(self._get_paren, expr)
+        # split
         expr = JSONPath.REP_DOUBLEDOT.sub(f"{JSONPath.SEP}..{JSONPath.SEP}", expr)
         expr = JSONPath.REP_DOT.sub(JSONPath.SEP, expr)
+        # put back
         expr = JSONPath.REP_PUT_PAREN.sub(self._put_paren, expr)
         expr = JSONPath.REP_PUT_BRACKET.sub(self._put_bracket, expr)
+        expr = JSONPath.REP_PUT_BACKQUOTE.sub(self._put_backquote, expr)
         expr = JSONPath.REP_PUT_QUOTE.sub(self._put_quote, expr)
         if expr.startswith("$;"):
             expr = expr[2:]
@@ -114,6 +127,7 @@ class JSONPath:
         logger.debug(f"after expr  : {expr}")
         return expr
 
+    # TODO abstract get and put procedures
     def _get_quote(self, m):
         n = len(self.subx["#Q"])
         self.subx["#Q"].append(m.group(1))
@@ -122,6 +136,14 @@ class JSONPath:
     def _put_quote(self, m):
         return self.subx["#Q"][int(m.group(1))]
 
+    def _get_backquote(self, m):
+        n = len(self.subx["#BQ"])
+        self.subx["#BQ"].append(m.group(1))
+        return f"`#BQ{n}`"
+
+    def _put_backquote(self, m):
+        return self.subx["#BQ"][int(m.group(1))]
+
     def _get_bracket(self, m):
         n = len(self.subx["#B"])
         self.subx["#B"].append(m.group(1))
@@ -139,7 +161,7 @@ class JSONPath:
         return self.subx["#P"][int(m.group(1))]
 
     @staticmethod
-    def _f_brackets(m):
+    def _gen_obj(m):
         ret = "__obj"
         for e in m.group(1).split("."):
             ret += '["%s"]' % e
@@ -155,7 +177,7 @@ class JSONPath:
                 f(v, i, f"{path}{JSONPath.SEP}{k}", *args)
 
     @staticmethod
-    def _getattr(obj: dict, path: str):
+    def _getattr(obj: dict, path: str, *, convert_number_str=False):
         r = obj
         for k in path.split("."):
             try:
@@ -163,7 +185,13 @@ class JSONPath:
             except (AttributeError, KeyError) as err:
                 logger.error(err)
                 return None
-
+        if convert_number_str and isinstance(r, str):
+            try:
+                if r.isdigit():
+                    return int(r)
+                return float(r)
+            except ValueError:
+                pass
         return r
 
     @staticmethod
@@ -171,15 +199,22 @@ class JSONPath:
         for sortby in sortbys.split(",")[::-1]:
             if sortby.startswith("~"):
                 obj.sort(
-                    key=lambda t, k=sortby: JSONPath._getattr(t[1], k[1:]), reverse=True
+                    key=lambda t, k=sortby: JSONPath._getattr(
+                        t[1], k[1:], convert_number_str=True
+                    ),
+                    reverse=True,
                 )
             else:
-                obj.sort(key=lambda t, k=sortby: JSONPath._getattr(t[1], k))
+                obj.sort(
+                    key=lambda t, k=sortby: JSONPath._getattr(
+                        t[1], k, convert_number_str=True
+                    )
+                )
 
     def _filter(self, obj, i: int, path: str, step: str):
         r = False
         try:
-            r = eval(step, None, {"__obj": obj})
+            r = self.eval_func(step, None, {"__obj": obj})
         except Exception as err:
             logger.error(err)
         if r:
@@ -230,7 +265,7 @@ class JSONPath:
         # slice
         if isinstance(obj, list) and JSONPath.REP_SLICE_CONTENT.fullmatch(step):
             obj = list(enumerate(obj))
-            vals = eval(f"obj[{step}]")
+            vals = self.eval_func(f"obj[{step}]")
             for idx, v in vals:
                 self._trace(v, i + 1, f"{path}{JSONPath.SEP}{idx}")
             return
@@ -245,7 +280,7 @@ class JSONPath:
         # filter
         if step.startswith("?(") and step.endswith(")"):
             step = step[2:-1]
-            step = JSONPath.REP_FILTER_CONTENT.sub(self._f_brackets, step)
+            step = JSONPath.REP_FILTER_CONTENT.sub(self._gen_obj, step)
             self._traverse(self._filter, obj, i + 1, path, step)
             return
 
@@ -278,6 +313,21 @@ class JSONPath:
             return
 
 
+def compile(expr):
+    return JSONPath(expr)
+
+
+# global cache
+_jsonpath_cache = {}
+
+
+def search(expr, data):
+    global _jsonpath_cache
+    if expr not in _jsonpath_cache:
+        _jsonpath_cache[expr] = JSONPath(expr)
+    return _jsonpath_cache[expr].parse(data)
+
+
 if __name__ == "__main__":
     with open("test/data/2.json", "rb") as f:
         d = json.load(f)
